// Oleg Petruny proprietary.


#include "MoveInteractableModificator.h"

#include "EnhancedInputComponent.h"
#include "InputMappingContext.h"

#include "Interactable/Interactable.h"
#include "PlayerBase.h"

UMoveInteractableModificator::UMoveInteractableModificator(const FObjectInitializer& ObjectInitializer)
    : UInteractableModificator(ObjectInitializer)
{
    activatorTypes |= static_cast<uint8>(EActivatorType::Use);

    inputMapping = LoadObject<UInputMappingContext>(this, TEXT("/Game/Input/Interactables/IMC_InteractableMove"));
    actor = Cast<AInteractable>(GetOwner());

    PrimaryComponentTick.bCanEverTick = true;
    PrimaryComponentTick.bStartWithTickEnabled = false;
}
void UMoveInteractableModificator::TickComponent(float DeltaTime, enum ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
    Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

    UpdatePosition(DeltaTime);
}


void UMoveInteractableModificator::Bind_Implementation(class UEnhancedInputComponent* input)
{
    if(!input && !inputMapping)
        return;

    if(state != State::Nothing)
        return;

    for(auto& mapping : inputMapping->GetMappings())
    {
        if(mapping.Key == EKeys::LeftMouseButton
            || mapping.Key == EKeys::RightMouseButton)
        {
            auto handle = input->BindAction(mapping.Action, ETriggerEvent::Started, this, &UMoveInteractableModificator::IncreaseState).GetHandle();
            bindindingHandlers.Add(handle);
            handle = input->BindAction(mapping.Action, ETriggerEvent::Completed, this, &UMoveInteractableModificator::DecreaseState).GetHandle();
            bindindingHandlers.Add(handle);
        }
        else if(mapping.Key == EKeys::MouseWheelAxis)
        {
            auto handle = input->BindAction(mapping.Action, ETriggerEvent::Triggered, this, &UMoveInteractableModificator::Zooming).GetHandle();
            bindindingHandlers.Add(handle);
        }
        else if(mapping.Key == EKeys::Mouse2D)
        {
            auto handle = input->BindAction(mapping.Action, ETriggerEvent::Triggered, this, &UMoveInteractableModificator::Rotating).GetHandle();
            bindindingHandlers.Add(handle);
        }
    }

    lastInput = input;
    player = Cast<APlayerBase>(Cast<APlayerController>(lastInput->GetOwner())->GetPawn());
    SetComponentTickEnabled(true);

    OnMoveActivated.Broadcast();
}

void UMoveInteractableModificator::Unbind_Implementation()
{
    if(!lastInput)
        return;

    if(state != State::Nothing)
        return;

    for(auto handle : bindindingHandlers)
    {
        lastInput->RemoveBindingByHandle(handle);
    }
    bindindingHandlers.Empty();

    state = static_cast<State>(0);
    ProcessState();
    SetComponentTickEnabled(false);
    OnMoveDeactivated.Broadcast();
}

void UMoveInteractableModificator::IncreaseState()
{
    state = static_cast<State>(static_cast<uint8>(state) + 1);
    if(static_cast<uint8>(state) > static_cast<uint8>(State::Rotating))
        state = State::Rotating;
    ProcessState();
}
void UMoveInteractableModificator::DecreaseState()
{
    state = static_cast<State>(static_cast<uint8>(state) - 1);
    if(static_cast<uint8>(state) > static_cast<uint8>(State::Rotating))
        state = State::Nothing;
    ProcessState();
}

void UMoveInteractableModificator::ProcessState()
{
    switch(state)
    {
        case State::Nothing:
        {
            player->cameraLocked = false;
            break;
        }
        case State::Holding:
        {
            player->cameraLocked = false;
            OnHolding.Broadcast();
            break;
        }
        case State::Rotating:
        {
            player->cameraLocked = true;
            OnRotating.Broadcast();
            break;
        }
        default:
            break;
    }
    if(state == State::Nothing)
    {
        if((actor->GetActivatedFlags() & static_cast<uint32>(EActivatorType::Use)) == 0
            && bindindingHandlers.Num() > 0)
        {
            Unbind_Implementation();
        }
    }
}

void UMoveInteractableModificator::UpdatePosition(float deltaTime)
{
    if(!player)
        return;
    if(state == State::Nothing)
    {
        distance = (player->GetCameraLocation() - actor->GetActorLocation()).Length();
        return;
    }
    auto camLoc = player->GetCameraLocation();
    auto dir = player->GetCameraDirection();
    auto newLoc = camLoc + dir * distance;
    auto currentLoc = actor->GetActorLocation();
    auto interpLoc = FMath::VInterpTo(currentLoc, newLoc, deltaTime, 20.0f);
    actor->SetActorLocation(interpLoc, true, nullptr, ETeleportType::None);
}

void UMoveInteractableModificator::Zooming(const FInputActionValue& value)
{
    if(state != State::Holding)
        return;

    distance += value.Get<float>() * zoomSpeed;
    distance = FMath::Clamp<float>(distance, minDistance, maxDistance);
}

void UMoveInteractableModificator::Rotating(const FInputActionValue& axis)
{
    if(state != State::Rotating)
        return;

    auto v = axis.Get<FVector>();
    auto vec = player->GetCameraDirection();
    actor->AddActorWorldRotation(FRotator{ v.Y * vec.X, v.X, v.Y * vec.Y * -1 }, true, nullptr, ETeleportType::None);
}
