// Oleg Petruny proprietary.


#include "PlayerBase.h"

#include "Camera/CameraComponent.h"
#include "EnhancedInputComponent.h"
#include "EnhancedInputSubsystems.h"
#include "GameFramework/CharacterMovementComponent.h"
#include "InputMappingContext.h"
#include "Kismet/GameplayStatics.h"
#include "Kismet/KismetMathLibrary.h"

#include "CustomGameInstanceBase.h"
#include "CustomGameUserSettings.h"
#include "Interactable/Activators/InteractableActivator.h"
#include "MainGameModeBase.h"

APlayerBase::APlayerBase()
{
    PrimaryActorTick.bCanEverTick = true;
}

void APlayerBase::Tick(float DeltaTime)
{
    ACharacter::Tick(DeltaTime);

    // stabilize move speed by fps
    AddActorLocalOffset(ConsumeMovementInputVector() * DeltaTime);

    if(moved)
    {
        moved = false;
        //OnPlayerMoved.Broadcast();
    }
}

void APlayerBase::SetupPlayerInputComponent(UInputComponent* PlayerInputComponent)
{
    ACharacter::SetupPlayerInputComponent(PlayerInputComponent);
}

void APlayerBase::BeginPlay()
{
    ACharacter::BeginPlay();

    auto world = GetWorld();

    GetCharacterMovement()->MaxWalkSpeed = moveSpeed;

    cameraManager = UGameplayStatics::GetPlayerCameraManager(world, 0);
    if(cameraManager)
    {
        cameraManager->ViewPitchMin = minPitch;
        cameraManager->ViewPitchMax = maxPitch;
    }

    auto gameSettings = UCustomGameUserSettings::GetCustomGameUserSettings();

    camera = FindComponentByClass<UCameraComponent>();
    if(camera)
    {
        camera->PostProcessSettings.MotionBlurAmount = gameSettings->bUseMotionBlur ? 1.0f : 0.0f;
    }


    if(auto PC = Cast<APlayerController>(GetController()))
    {
        inputComponent = Cast<UEnhancedInputComponent>(PC->InputComponent);;

        if(auto inputSubsystem = ULocalPlayer::GetSubsystem<UEnhancedInputLocalPlayerSubsystem>(PC->GetLocalPlayer()))
        {
            if(auto GI = Cast<UCustomGameInstanceBase>(GetWorld()->GetGameInstance()))
            {
                inputSubsystem->ClearAllMappings();
                for(auto& inputContext : GI->inputContexts)
                {
                    inputSubsystem->AddMappingContext(inputContext.LoadSynchronous(), 0);
                }
            }
        }
    }

    LoadInteractablesActivators();
}

APlayerBase* APlayerBase::LockCamera(UWorld* world, bool locked)
{
    if(auto PC = UGameplayStatics::GetPlayerController(world, 0))
    {
        if(auto pawn = Cast<APlayerBase>(PC->GetPawn()))
        {
            pawn->cameraLocked = locked;
            return pawn;
        }
    }

    return nullptr;
}

void APlayerBase::AttachToCamera(AActor* actor)
{
    if(!actor)
        return;

    actor->AttachToComponent(camera, FAttachmentTransformRules::KeepWorldTransform);
}

FVector APlayerBase::GetCameraDirection()
{
    return camera->GetComponentRotation().Vector();
}
FVector APlayerBase::GetCameraLocation()
{
    return camera->GetComponentLocation();
}

void APlayerBase::SwitchToCameraPawn()
{
    if(auto gamemode_base = UGameplayStatics::GetGameMode(GetWorld()))
    {
        if(auto gamemode = Cast<AMainGameModeBase>(gamemode_base))
        {
            gamemode->SwitchCameraMode();
        }
    }
}

void APlayerBase::MoveCamera(FVector2D value)
{
    if(cameraLocked)
        return;

    AddControllerYawInput(value.X);
    AddControllerPitchInput(value.Y);

    moved = true;
}

void APlayerBase::MoveCharacter(FVector2D value)
{
    if(moveLocked)
        return;

    AddMovementInput(GetActorRightVector(), value.X);
    AddMovementInput(GetActorForwardVector(), value.Y);

    moved = true;
}

void APlayerBase::Jump()
{
    if(jumpLocked)
        return;

    ACharacter::Jump();

    moved = true;
}

void APlayerBase::SwitchRun(bool run)
{
    if(runLocked)
        return;

    if(run)
        GetCharacterMovement()->MaxWalkSpeed = moveSpeed * runSpeedMultiplier;
    else
        GetCharacterMovement()->MaxWalkSpeed = moveSpeed;
}

void APlayerBase::UpdatePitch(float min, float max)
{
    minPitch = min;
    maxPitch = max;

    if(cameraManager)
    {
        cameraManager->ViewPitchMin = minPitch;
        cameraManager->ViewPitchMax = maxPitch;
    }
}

void APlayerBase::LoadInteractablesActivators()
{
    if(auto GI = Cast<UCustomGameInstanceBase>(GetWorld()->GetGameInstance()))
    {
        TSet<UClass*> instancedActivators;
        for(auto& act : GI->interactionsActivators)
        {
            if(instancedActivators.Contains(act))
                continue;
            instancedActivators.Add(act);

            auto component = NewObject<UInteractableActivator>(this, act);
            component->interactableActivatedDelegate.BindUObject(this, &APlayerBase::InteractableActivated);
            component->interactableDeactivatedDelegate.BindUObject(this, &APlayerBase::InteractableDeactivated);
            component->SetupAttachment(camera);
            component->RegisterComponent();
        }
    }
}

void APlayerBase::InteractableActivated(AInteractable* interactable, EActivatorType type)
{
    if(!interactable)
        return;

    if(interactionLocked)
        return;

    interactable->_Activate(type);
}

void APlayerBase::InteractableDeactivated(AInteractable* interactable, EActivatorType type)
{
    if(!interactable)
        return;

    if(interactionLocked)
        return;

    interactable->_Deactivate(type);
}